import fetch_csv_from_compass
import merge_local_csv
import deskewer
import ais_to_csai_csv_converter
import postResults
from importlib import reload
import find_archetypes
import os
import glob
import pandas as pd
import json
import shutil
import plotly.express as px
archetypes = reload(find_archetypes)

  # To use the iNav-bundled Neo4J instance, use "neo4j:7687", username neo4j:1234
NEO4J_URI = "NA"
NEO4J_USERNAME = "NA"
NEO4J_PASSWORD = "NA"

# Feature Flags - change these to turn on or off specific functionality in future notebook cells
USE_CSV_DATASET = True
CSV_DATASET = "/data/deskewed.csv"

CATEGORY_BOOST = 1.5
TAG_BOOST = 1.5

USE_NEO4J_DATASET = False

USE_NEO4J_IMPORTS = False
USE_CSA_FINDINGS_IMPORTS = False
USE_CSA_FINDINGS_INVOCATIONS = False
USE_CSA_FINDINGS_ANNOTATIONS = False
USE_NEO4J_INVOCATIONS = False
USE_CSA_TAGS = False
USE_CATEGORIES=False

CALCULATE_CSA_SCORE=False
USE_DESKEW_DATA = True
prefix = ""

COMPASS_CALLBACK_URL = os.getenv("PORTFOLIO_API_URL")

analyzer = archetypes.DataAnalyzer(
    neo4jHost = NEO4J_URI,
    neo4User = NEO4J_USERNAME,
    neo4jPassword = NEO4J_PASSWORD,
    outputDirectory = '/data/outputs/',

    # Feature Flags
    useCsvDataSet = USE_CSV_DATASET,
    csvDataSet = CSV_DATASET,
    categoryBoost = CATEGORY_BOOST,
    tagBoost = TAG_BOOST,

    useNeo4jDataSet = USE_NEO4J_DATASET,
    useNeo4jImports = USE_NEO4J_IMPORTS,
    useNeo4jCsaFindingsImports = USE_CSA_FINDINGS_IMPORTS,
    useNeo4jCsaFindingsInvocations = USE_CSA_FINDINGS_INVOCATIONS,
    useNeo4jCsaFindingsAnnotations = USE_CSA_FINDINGS_ANNOTATIONS,
    useNeo4jInvocations = USE_NEO4J_INVOCATIONS,
    useCsaTags = USE_CSA_TAGS,
    useCategories = USE_CATEGORIES,
    calculateCsaScores = CALCULATE_CSA_SCORE,
    deskewData = USE_DESKEW_DATA,
)

def run(groupFilter, clusterPrefix, metrics, epochs, samples, n_neighbors, min_dists, min_clusters ):
    print ('====> Run')
    prefix = clusterPrefix
    
    if COMPASS_CALLBACK_URL:
        # Clean up existing files
        for root, dirs, files in os.walk('/data'):
            for f in files:
                os.unlink(os.path.join(root, f))
        for d in dirs:
            shutil.rmtree(os.path.join(root, d))
        print ('====> Fetch Data From Compass')
        fetch_csv_from_compass.fetchData(groupFilter)
    else:
        print ('====> Merge Local CSVs')
        merge_local_csv.merge()

    print ('====> Convert Data')
    ais_to_csai_csv_converter.convert()

    print ('====> Deskew Data')
    deskewer.deskew("/data/skewed.csv", 'total', 2)
    
    print ('====> Clean up Output Dir')
    # Clean previous output if any
    for root, dirs, files in os.walk('/data/outputs/'):
        for f in files:
            os.unlink(os.path.join(root, f))
    for d in dirs:
        shutil.rmtree(os.path.join(root, d))

    def add_score(appName, score):
        return f"{appName}[{score}]"

    consolidatedDf, csaScoreDf = analyzer.costruct_facets_dataframe()
    print ("====> ANALYZER")
    #print (consolidatedDf)

    #print("Shape of your dataset:")
    #print(consolidatedDf.shape)

    ##print("\nTotal applications present in data set:")
    ##print(consolidatedDf['app'].drop_duplicates())

    #print("\nTop ten entries in your data set:")
    #print(consolidatedDf.sample(n=10))

    #print(consolidatedDf.sort_values(by= 'total', ascending = False))

    #consolidatedDf = consolidatedDf[consolidatedDf["total"] > 1.5]
    consolidatedDf.hist()

    consolidatedDf.sort_values('total')

    consolidatedDf=consolidatedDf.round({'total': 3})

    consolidatedDf.drop_duplicates(inplace=True)

    tmp = consolidatedDf.copy()

    # --- Sort data and create pivot table. This will be a sparsely populated table.
    tmp.sort_values(by=['app', 'facet',"total"], inplace=True)
    #tmp.to_csv ("/data/test.csv")
    # --- convert dataframe into a pivot table, with each unique package in its own column
    pivotDf = tmp.pivot_table('total', index='app', columns = ['facet'], fill_value=0)

    # --- Convert sparsely populated table into numpy array
    # --- preserve the appNames column to be added back in in a later stage
    appNames = pd.DataFrame(pivotDf.index)
    print ("-----DISCOVER CLUSTERS------")
    umap_results = analyzer.discover_archetypes(consolidatedDf,metrics,epochs,samples,n_neighbors,min_dists, min_clusters)

    #print (umap_results)

    sorted_results = sorted(umap_results.items(), key=lambda i: sort_umap_results(i[0], i[1]), reverse=True)
    render_result(sorted_results[0][1], prefix)

def explain_results(outputs, prefix):
    print ("------EXPLAIN RESULTS--------")
    # unpack the data we have to work with
    pivotDf, result, fig = outputs
    sk_score, cluster_count, unassigned, reducedEmbedded, labels, cluster_list = result
    #n_neighbors, min_dist, n_components, metric, samples, epoch_cnt, run_count, min_cluster = params
    
    appNames = pd.DataFrame(pivotDf.index)
    clustersDf = pd.DataFrame(reducedEmbedded, columns=[])
    clustersDf["app"] = appNames
    clustersDf['cluster'] = labels.tolist()
    cluster_column = clustersDf.pop('cluster')
    
    facetsDf = pivotDf.copy()
    clustersDf = pd.merge(facetsDf, clustersDf, on='app', how='left')
    
    if CALCULATE_CSA_SCORE:
        clustersDf = pd.merge(clustersDf, csaScoreDf, on='app',how='left')
        clustersDf['score'] = clustersDf['score'].fillna(10)
        clustersDf["app"] = clustersDf.apply(lambda row: add_score(row.app,row.score), axis=1)
        clustersDf.drop("score", inplace=True, axis=1)
    
    clustersDf.insert(1, 'clusters', cluster_column)
    clustersDf = clustersDf.set_index('app')
    clustersDf = clustersDf.sort_values(by=['clusters'])
    
    print ('====> Post Artifact')
    postResults.postArtifact(f"{COMPASS_CALLBACK_URL}/artifact",f"/data/outputs/{analyzer.latestFile}", "CSAI 3D Cluster", "CSAI generated cluster for ["+prefix+"]", "CSAI Plugin", "CSAI" ) 
    print ('====> Post Artifact: Done!') 

    # for each cluster, filter and sort to show app, cluster, and top X facets
    for c in clustersDf['clusters'].unique():
        print("Cluster {}".format(c))
        curr_clusterDf = clustersDf.loc[clustersDf['clusters'] == c]
        
        
        curr_cluster_apps = curr_clusterDf.iloc[: , :1] #

        # sort the current cluster's facets so the most common occur in earlier columns
        curr_cluster_facets = curr_clusterDf.iloc[: , 1:] #
        
        curr_cluster_facets_sums = curr_cluster_facets.sum() # 
        curr_cluster_facets_sums = curr_cluster_facets_sums.head(curr_cluster_facets_sums.shape[0] -1)
        
        # display(curr_cluster_facets_sums)
        curr_cluster_facets_count = len(pivotDf.columns)
        curr_cluster_facets_sums.drop(curr_cluster_facets_sums.tail(1).index,inplace=True) # drop last n rows

        #curr_cluster_facets_count = curr_cluster_facets_count.head(curr_cluster_facets_count.shape[0] -1)
        curr_cluster_facets_sums.to_csv("/data/outputs/clusters_"+"Cluster {}".format(c)+".csv")
        
        curr_cluster_facets=curr_cluster_facets[curr_cluster_facets_sums.sort_values(ascending=False).index[:curr_cluster_facets_count-1]]

        
        # build a table showing this cluster's top X facets
        columnsToKeep = 10
        curr_clusters_top_facets = pd.concat([
            curr_cluster_apps,
            curr_cluster_facets.iloc[: , :columnsToKeep]
        ], axis=1)

        heatMapDf = curr_clusters_top_facets.drop("clusters", axis=1)
        heatmap_height = 250 + (10*2.3*heatMapDf.shape[0])
        heatmap_width = 50*2.3*heatMapDf.shape[1]
        # display("height: " + str(heatmap_height))
        # display("width: " + str(heatmap_width))
        fig = px.imshow(heatMapDf, width=heatmap_width, height=heatmap_height)
        fig.update_layout(
        showlegend = False,
        width=heatmap_width,
        height=heatmap_height,
        autosize = False)
        fig.update_xaxes(side="top", automargin=True)
        fig.update_yaxes(automargin=True)
        #print (prefix)
        clusterFileName = f"/data/outputs/archetype_clusters_{prefix}"+"_Cluster {}".format(c)+".html"
        fig.write_html(clusterFileName)
        #print(clusterFileName)
        curr_cluster_apps.reset_index(inplace=True)
        clusterName = f"{prefix}"+" Cluster {}".format(c)
        #print(clusterName)

        if COMPASS_CALLBACK_URL:
            for ind in curr_cluster_apps.index:
                appName = curr_cluster_apps['app'][ind]
                print(f"Storing records for {appName}")
                postResults.postRecord(f"{COMPASS_CALLBACK_URL}/applicationRepo/genericRecord", "CSAI",appName, f"/data/outputs/{analyzer.latestFile}",  clusterName) 
                postResults.postFile(f"{COMPASS_CALLBACK_URL}/applicationRepo/byName/{appName}/file/upload", clusterFileName,  f"CSAI Cluster {clusterName} app to facet heat map") 
                postResults.postProperty(f"{COMPASS_CALLBACK_URL}/applicationRepo/byName/{appName}/fields", "clusterName", clusterName)
                
         

def sort_umap_results(params, outputs):
    pivotDf, result, fig = outputs
    return result[0]

def result_label(params, outputs):
    pivotDf, result, fig = outputs
    return "Score: {}, Clusters: {}, Unassigned: {}".format(result[0], result[1], result[2])
    #results_options = [(result_label(i[0], i[1]), i[1]) for i in sorted_results]

def render_result(outputs, prefix):
    pivotDf, result, fig = outputs
    sk_score, cluster_count, unassigned, reducedEmbedded, labels, cluster_list = result
    print("Clustered {} apps into {} clusters.  {} apps are unassigned to a cluster.".format(pivotDf.shape[0], result[1], result[2]))
    explain_results(outputs, prefix)
